@@ -1,4 +1,3 @@
-# -*- coding: utf-8; fill-column: 76 -*-
 """Signals and events.
 
 A small implementation of signals, inspired by a snippet of Django signal
@@ -8,26 +7,40 @@ each manages its own receivers and message emission.
 The :func:`signal` function provides singleton behavior for named signals.
 
 """
+from __future__ import annotations
+
+import typing as t
+from collections import defaultdict
 from contextlib import contextmanager
 from warnings import warn
 from weakref import WeakValueDictionary
 
-from blinker._utilities import (
-    WeakTypes,
-    defaultdict,
-    hashable_identity,
-    lazy_property,
-    reference,
-    symbol,
-    )
+from blinker._utilities import annotatable_weakref
+from blinker._utilities import hashable_identity
+from blinker._utilities import IdentityType
+from blinker._utilities import is_coroutine_function
+from blinker._utilities import lazy_property
+from blinker._utilities import reference
+from blinker._utilities import symbol
+from blinker._utilities import WeakTypes
 
+if t.TYPE_CHECKING:
+    import typing_extensions as te
 
-ANY = symbol('ANY')
+    T_callable = t.TypeVar("T_callable", bound=t.Callable[..., t.Any])
+
+    T = t.TypeVar("T")
+    P = te.ParamSpec("P")
+
+    AsyncWrapperType = t.Callable[[t.Callable[P, t.Awaitable[T]]], t.Callable[P, T]]
+    SyncWrapperType = t.Callable[[t.Callable[P, T]], t.Callable[P, t.Awaitable[T]]]
+
+ANY = symbol("ANY")
 ANY.__doc__ = 'Token for "any sender".'
 ANY_ID = 0
 
 
-class Signal(object):
+class Signal:
     """A notification emitter."""
 
     #: An :obj:`ANY` convenience synonym, allows ``Signal.ANY``
@@ -35,7 +48,7 @@ class Signal(object):
     ANY = ANY
 
     @lazy_property
-    def receiver_connected(self):
+    def receiver_connected(self) -> Signal:
         """Emitted after each :meth:`connect`.
 
         The signal sender is the signal instance, and the :meth:`connect`
@@ -47,7 +60,7 @@ class Signal(object):
         return Signal(doc="Emitted after a receiver connects.")
 
     @lazy_property
-    def receiver_disconnected(self):
+    def receiver_disconnected(self) -> Signal:
         """Emitted after :meth:`disconnect`.
 
         The sender is the signal instance, and the :meth:`disconnect` arguments
@@ -70,7 +83,7 @@ class Signal(object):
         """
         return Signal(doc="Emitted after a receiver disconnects.")
 
-    def __init__(self, doc=None):
+    def __init__(self, doc: str | None = None) -> None:
         """
         :param doc: optional.  If provided, will be assigned to the signal's
           __doc__ attribute.
@@ -84,12 +97,15 @@ class Signal(object):
         #: internal :class:`Signal` implementation, however the boolean value
         #: of the mapping is useful as an extremely efficient check to see if
         #: any receivers are connected to the signal.
-        self.receivers = {}
-        self._by_receiver = defaultdict(set)
-        self._by_sender = defaultdict(set)
-        self._weak_senders = {}
+        self.receivers: dict[IdentityType, t.Callable | annotatable_weakref] = {}
+        self.is_muted = False
+        self._by_receiver: dict[IdentityType, set[IdentityType]] = defaultdict(set)
+        self._by_sender: dict[IdentityType, set[IdentityType]] = defaultdict(set)
+        self._weak_senders: dict[IdentityType, annotatable_weakref] = {}
 
-    def connect(self, receiver, sender=ANY, weak=True):
+    def connect(
+        self, receiver: T_callable, sender: t.Any = ANY, weak: bool = True
+    ) -> T_callable:
         """Connect *receiver* to signal events sent by *sender*.
 
         :param receiver: A callable.  Will be invoked by :meth:`send` with
@@ -109,11 +125,14 @@ class Signal(object):
 
         """
         receiver_id = hashable_identity(receiver)
+        receiver_ref: T_callable | annotatable_weakref
+
         if weak:
             receiver_ref = reference(receiver, self._cleanup_receiver)
             receiver_ref.receiver_id = receiver_id
         else:
             receiver_ref = receiver
+        sender_id: IdentityType
         if sender is ANY:
             sender_id = ANY_ID
         else:
@@ -136,28 +155,27 @@ class Signal(object):
                 del sender_ref
 
         # broadcast this connection.  if receivers raise, disconnect.
-        if ('receiver_connected' in self.__dict__ and
-            self.receiver_connected.receivers):
+        if "receiver_connected" in self.__dict__ and self.receiver_connected.receivers:
             try:
-                self.receiver_connected.send(self,
-                                             receiver=receiver,
-                                             sender=sender,
-                                             weak=weak)
-            except:
+                self.receiver_connected.send(
+                    self, receiver=receiver, sender=sender, weak=weak
+                )
+            except TypeError as e:
                 self.disconnect(receiver, sender)
-                raise
+                raise e
         if receiver_connected.receivers and self is not receiver_connected:
             try:
-                receiver_connected.send(self,
-                                        receiver_arg=receiver,
-                                        sender_arg=sender,
-                                        weak_arg=weak)
-            except:
+                receiver_connected.send(
+                    self, receiver_arg=receiver, sender_arg=sender, weak_arg=weak
+                )
+            except TypeError as e:
                 self.disconnect(receiver, sender)
-                raise
+                raise e
         return receiver
 
-    def connect_via(self, sender, weak=False):
+    def connect_via(
+        self, sender: t.Any, weak: bool = False
+    ) -> t.Callable[[T_callable], T_callable]:
         """Connect the decorated function as a receiver for *sender*.
 
         :param sender: Any object or :obj:`ANY`.  The decorated function
@@ -178,13 +196,17 @@ class Signal(object):
         .. versionadded:: 1.1
 
         """
-        def decorator(fn):
+
+        def decorator(fn: T_callable) -> T_callable:
             self.connect(fn, sender, weak)
             return fn
+
         return decorator
 
     @contextmanager
-    def connected_to(self, receiver, sender=ANY):
+    def connected_to(
+        self, receiver: t.Callable, sender: t.Any = ANY
+    ) -> t.Generator[None, None, None]:
         """Execute a block with the signal temporarily connected to *receiver*.
 
         :param receiver: a receiver callable
@@ -207,13 +229,25 @@ class Signal(object):
         self.connect(receiver, sender=sender, weak=False)
         try:
             yield None
-        except:
-            self.disconnect(receiver)
-            raise
-        else:
+        finally:
             self.disconnect(receiver)
 
-    def temporarily_connected_to(self, receiver, sender=ANY):
+    @contextmanager
+    def muted(self) -> t.Generator[None, None, None]:
+        """Context manager for temporarily disabling signal.
+        Useful for test purposes.
+        """
+        self.is_muted = True
+        try:
+            yield None
+        except Exception as e:
+            raise e
+        finally:
+            self.is_muted = False
+
+    def temporarily_connected_to(
+        self, receiver: t.Callable, sender: t.Any = ANY
+    ) -> t.ContextManager[None]:
         """An alias for :meth:`connected_to`.
 
         :param receiver: a receiver callable
@@ -226,12 +260,18 @@ class Signal(object):
           deprecated in 1.2 and will be removed in a subsequent version.
 
         """
-        warn("temporarily_connected_to is deprecated; "
-             "use connected_to instead.",
-             DeprecationWarning)
+        warn(
+            "temporarily_connected_to is deprecated; use connected_to instead.",
+            DeprecationWarning,
+        )
         return self.connected_to(receiver, sender)
 
-    def send(self, *sender, **kwargs):
+    def send(
+        self,
+        *sender: t.Any,
+        _async_wrapper: AsyncWrapperType | None = None,
+        **kwargs: t.Any,
+    ) -> list[tuple[t.Callable, t.Any]]:
         """Emit this signal on behalf of *sender*, passing on ``kwargs``.
 
         Returns a list of 2-tuples, pairing receivers with their return
@@ -239,15 +279,65 @@ class Signal(object):
 
         :param sender: Any object or ``None``.  If omitted, synonymous
           with ``None``.  Only accepts one positional argument.
+        :param _async_wrapper: A callable that should wrap a coroutine
+          receiver and run it when called synchronously.
 
         :param kwargs: Data to be sent to receivers.
         """
+        if self.is_muted:
+            return []
+
+        sender = self._extract_sender(sender)
+        results = []
+        for receiver in self.receivers_for(sender):
+            if is_coroutine_function(receiver):
+                if _async_wrapper is None:
+                    raise RuntimeError("Cannot send to a coroutine function")
+                receiver = _async_wrapper(receiver)
+            result = receiver(sender, **kwargs)
+            results.append((receiver, result))
+        return results
+
+    async def send_async(
+        self,
+        *sender: t.Any,
+        _sync_wrapper: SyncWrapperType | None = None,
+        **kwargs: t.Any,
+    ) -> list[tuple[t.Callable, t.Any]]:
+        """Emit this signal on behalf of *sender*, passing on ``kwargs``.
+
+        Returns a list of 2-tuples, pairing receivers with their return
+        value. The ordering of receiver notification is undefined.
+
+        :param sender: Any object or ``None``.  If omitted, synonymous
+          with ``None``. Only accepts one positional argument.
+        :param _sync_wrapper: A callable that should wrap a synchronous
+          receiver and run it when awaited.
+
+        :param kwargs: Data to be sent to receivers.
+        """
+        if self.is_muted:
+            return []
+
+        sender = self._extract_sender(sender)
+        results = []
+        for receiver in self.receivers_for(sender):
+            if not is_coroutine_function(receiver):
+                if _sync_wrapper is None:
+                    raise RuntimeError("Cannot send to a non-coroutine function")
+                receiver = _sync_wrapper(receiver)
+            result = await receiver(sender, **kwargs)
+            results.append((receiver, result))
+        return results
+
+    def _extract_sender(self, sender: t.Any) -> t.Any:
         if not self.receivers:
             # Ensure correct signature even on no-op sends, disable with -O
             # for lowest possible cost.
             if __debug__ and sender and len(sender) > 1:
-                raise TypeError('send() accepts only one positional '
-                                'argument, %s given' % len(sender))
+                raise TypeError(
+                    f"send() accepts only one positional argument, {len(sender)} given"
+                )
             return []
 
         # Using '*sender' rather than 'sender=None' allows 'sender' to be
@@ -256,14 +346,14 @@ class Signal(object):
         if len(sender) == 0:
             sender = None
         elif len(sender) > 1:
-            raise TypeError('send() accepts only one positional argument, '
-                            '%s given' % len(sender))
+            raise TypeError(
+                f"send() accepts only one positional argument, {len(sender)} given"
+            )
         else:
             sender = sender[0]
-        return [(receiver, receiver(sender, **kwargs))
-                for receiver in self.receivers_for(sender)]
+        return sender
 
-    def has_receivers_for(self, sender):
+    def has_receivers_for(self, sender: t.Any) -> bool:
         """True if there is probably a receiver for *sender*.
 
         Performs an optimistic check only.  Does not guarantee that all
@@ -279,14 +369,15 @@ class Signal(object):
             return False
         return hashable_identity(sender) in self._by_sender
 
-    def receivers_for(self, sender):
+    def receivers_for(
+        self, sender: t.Any
+    ) -> t.Generator[t.Callable[[t.Any], t.Any], None, None]:
         """Iterate all live receivers listening for *sender*."""
         # TODO: test receivers_for(ANY)
         if self.receivers:
             sender_id = hashable_identity(sender)
             if sender_id in self._by_sender:
-                ids = (self._by_sender[ANY_ID] |
-                       self._by_sender[sender_id])
+                ids = self._by_sender[ANY_ID] | self._by_sender[sender_id]
             else:
                 ids = self._by_sender[ANY_ID].copy()
             for receiver_id in ids:
@@ -299,9 +390,9 @@ class Signal(object):
                         self._disconnect(receiver_id, ANY_ID)
                         continue
                     receiver = strong
-                yield receiver
+                yield receiver  # type: ignore[misc]
 
-    def disconnect(self, receiver, sender=ANY):
+    def disconnect(self, receiver: t.Callable, sender: t.Any = ANY) -> None:
         """Disconnect *receiver* from this signal's events.
 
         :param receiver: a previously :meth:`connected<connect>` callable
@@ -310,6 +401,7 @@ class Signal(object):
           to disconnect from all senders.  Defaults to ``ANY``.
 
         """
+        sender_id: IdentityType
         if sender is ANY:
             sender_id = ANY_ID
         else:
@@ -317,13 +409,13 @@ class Signal(object):
         receiver_id = hashable_identity(receiver)
         self._disconnect(receiver_id, sender_id)
 
-        if ('receiver_disconnected' in self.__dict__ and
-            self.receiver_disconnected.receivers):
-            self.receiver_disconnected.send(self,
-                                            receiver=receiver,
-                                            sender=sender)
+        if (
+            "receiver_disconnected" in self.__dict__
+            and self.receiver_disconnected.receivers
+        ):
+            self.receiver_disconnected.send(self, receiver=receiver, sender=sender)
 
-    def _disconnect(self, receiver_id, sender_id):
+    def _disconnect(self, receiver_id: IdentityType, sender_id: IdentityType) -> None:
         if sender_id == ANY_ID:
             if self._by_receiver.pop(receiver_id, False):
                 for bucket in self._by_sender.values():
@@ -333,19 +425,19 @@ class Signal(object):
             self._by_sender[sender_id].discard(receiver_id)
             self._by_receiver[receiver_id].discard(sender_id)
 
-    def _cleanup_receiver(self, receiver_ref):
+    def _cleanup_receiver(self, receiver_ref: annotatable_weakref) -> None:
         """Disconnect a receiver from all senders."""
-        self._disconnect(receiver_ref.receiver_id, ANY_ID)
+        self._disconnect(t.cast(IdentityType, receiver_ref.receiver_id), ANY_ID)
 
-    def _cleanup_sender(self, sender_ref):
+    def _cleanup_sender(self, sender_ref: annotatable_weakref) -> None:
         """Disconnect all receivers from a sender."""
-        sender_id = sender_ref.sender_id
+        sender_id = t.cast(IdentityType, sender_ref.sender_id)
         assert sender_id != ANY_ID
         self._weak_senders.pop(sender_id, None)
         for receiver_id in self._by_sender.pop(sender_id, ()):
             self._by_receiver[receiver_id].discard(sender_id)
 
-    def _cleanup_bookkeeping(self):
+    def _cleanup_bookkeeping(self) -> None:
         """Prune unused sender/receiver bookkeeping. Not threadsafe.
 
         Connecting & disconnecting leave behind a small amount of bookkeeping
@@ -371,7 +463,7 @@ class Signal(object):
                 if not bucket:
                     mapping.pop(_id, None)
 
-    def _clear_state(self):
+    def _clear_state(self) -> None:
         """Throw away all signal state.  Useful for unit tests."""
         self._weak_senders.clear()
         self.receivers.clear()
@@ -379,7 +471,8 @@ class Signal(object):
         self._by_receiver.clear()
 
 
-receiver_connected = Signal("""\
+receiver_connected = Signal(
+    """\
 Sent by a :class:`Signal` after a receiver connects.
 
 :argument: the Signal that was connected to
@@ -394,36 +487,38 @@ As of 1.2, individual signals have their own private
 :attr:`~Signal.receiver_disconnected` signals with a slightly simplified
 call signature.  This global signal is planned to be removed in 1.6.
 
-""")
+"""
+)
 
 
 class NamedSignal(Signal):
     """A named generic notification emitter."""
 
-    def __init__(self, name, doc=None):
+    def __init__(self, name: str, doc: str | None = None) -> None:
         Signal.__init__(self, doc)
 
         #: The name of this signal.
         self.name = name
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         base = Signal.__repr__(self)
-        return "{}; {!r}>".format(base[:-1], self.name)
+        return f"{base[:-1]}; {self.name!r}>"
 
 
 class Namespace(dict):
     """A mapping of signal names to signals."""
 
-    def signal(self, name, doc=None):
+    def signal(self, name: str, doc: str | None = None) -> NamedSignal:
         """Return the :class:`NamedSignal` *name*, creating it if required.
 
         Repeated calls to this function will return the same signal object.
 
         """
         try:
-            return self[name]
+            return self[name]  # type: ignore[no-any-return]
         except KeyError:
-            return self.setdefault(name, NamedSignal(name, doc))
+            result = self.setdefault(name, NamedSignal(name, doc))
+            return result  # type: ignore[no-any-return]
 
 
 class WeakNamespace(WeakValueDictionary):
@@ -437,16 +532,17 @@ class WeakNamespace(WeakValueDictionary):
 
     """
 
-    def signal(self, name, doc=None):
+    def signal(self, name: str, doc: str | None = None) -> NamedSignal:
         """Return the :class:`NamedSignal` *name*, creating it if required.
 
         Repeated calls to this function will return the same signal object.
 
         """
         try:
-            return self[name]
+            return self[name]  # type: ignore[no-any-return]
         except KeyError:
-            return self.setdefault(name, NamedSignal(name, doc))
+            result = self.setdefault(name, NamedSignal(name, doc))
+            return result  # type: ignore[no-any-return]
 
 
 signal = Namespace().signal
